char rcsid_parse_command[] = "$Id: parse_command.c,v 1.3 1992/08/10 17:35:34 pjames Exp $";

#include "my-stdio.h"
#include "my-stdlib.h"
#include "my-string.h"
#include "my-time.h"

#include "config.h"
#include "list.h"
#include "match.h"
#include "parse_command.h"
#include "storage.h"
#include "structures.h"
#include "utils.h"

static void
parse_into_words(char *input, int *argc, char *argv[])
{
    int		in_quotes = 0;
    char       *ptr = input;

    while (*input == ' ')
	input++;

    for (*argc = 0; *argc < MAXARGS  &&  *input != '\0'; (*argc)++) {
	argv[*argc] = ptr;
	while (*input != '\0'  &&  (in_quotes  ||  *input != ' ')) {
	    char	c = *(input++);

	    if (c == '"')
		in_quotes = !in_quotes;
	    else if (c == '\\') {
		if (*input != '\0')
		    *(ptr++) = *(input++);
	    } else
		*(ptr++) = c;
	}
	while (*input == ' ')
	    input++;
	*(ptr++) = '\0';
    }
}

static char *
build_string(int argc, char *argv[])
{
    int         i, len = 0;
    char	*str;

    if (!argc)
        return strdup("");

    len = strlen(argv[0]);
    for (i = 1; i < argc; i++)
	len += 1 + strlen(argv[i]);

    str = (char *) mymalloc(len + 1, M_STRING);

    strcpy(str, argv[0]);
    for (i = 1; i < argc; i++) {
	strcat(str, " ");
	strcat(str, argv[i]);
    }

    return str;
}

/* Prepositions */

#define MAXPPHRASE 5	/* max. number of words in a prepositional phrase */
static const char  *prep_list[] = {
    "with/using",
    "at/to",
    "in front of",
    "in/inside/into",
    "on top of/on/onto/upon",
    "out of/from inside/from",
    "over",
    "through",
    "under/underneath/beneath",
    "behind",
    "beside",
    "for/about",
    "is",
    "as",
    "off/off of",
};

#define NPREPS Arraysize(prep_list, const char *)

struct  pt_entry {
    int		nwords;
    char	*words[MAXPPHRASE];
    struct pt_entry *next;
};

struct pt_entry	*prep_table[NPREPS];

void
build_preptable(void)
{
    int		i, j;
    int		wordc;
    char	*words[MAXPPHRASE];
    char	cprep[100];
    const char	*p;
    char	*t;
    struct pt_entry	*current_alias, **prev;

    for (i = 0; i < NPREPS; i++) {
	p = prep_list[i];
	prev = &(prep_table[i]);
	while (*p) {
	    t = cprep;
	    if (*p == '/')
		p++;
	    while (*p && *p != '/')
		*t++ = *p++;
	    *t = '\0';

	    parse_into_words(cprep, &wordc, words);

	    current_alias = (struct pt_entry *)
			    mymalloc(sizeof(struct pt_entry), M_PREP);
	    current_alias->nwords = wordc;
	    current_alias->next = 0;
	    for (j = 0; j < wordc; j++)
		current_alias->words[j] = strdup(words[j]);
	    if (prev) 
		*prev = current_alias;
	    prev = &current_alias->next;
	}
    }
}

static Prepid
prep_find(int argc, char *argv[], int *pstart, int *pend)
{
    struct pt_entry	*alias;
    int			i, j, k;
    int			exact_match = (pstart == 0  ||  pend == 0);

    for (i = 0; i < argc; i++) {
	for (j = 0; j < NPREPS; j++) {
	    for (alias = prep_table[j]; alias; alias = alias->next) {
		if (i + alias->nwords <= argc) {
		    for (k = 0; k < alias->nwords; k++) {
			if (mystrcasecmp(argv[i + k], alias->words[k]))
			    break;
		    }
		    if (k == alias->nwords
			     &&  (! exact_match || i + k == argc)) {
			if (! exact_match) {
			    *pstart = i;
			    *pend = i + alias->nwords - 1;
			}
			return j;
		    }
		}
	    }
	}
	if (exact_match)
	    break;
    }
    return FAILED_MATCH;
}

Prepid
match_prep(char *prepname)
{
    Prepid	prep;
    int		argc;
    char	*ptr;
    char	*argv[MAXARGS];

    if (!mystrcasecmp(prepname, "none")) return NOTHING;
    if (!mystrcasecmp(prepname, "any")) return AMBIGUOUS;

    if (prepname[0] == '#')
	prepname++;
    prep = strtol(prepname, &ptr, 10);
    if (*ptr == '\0') {
	if (prep < 0 || prep >= NPREPS || (prep == 0 && prepname[0] != '0'))
	    return FAILED_MATCH;
	else
	    return prep;
    }

    if (ptr = strchr(prepname, '/'))
	*ptr = '\0';

    parse_into_words(prepname, &argc, argv);
    return prep_find(argc, argv, 0, 0);
}

const char *
prep_name(Prepid prep)
{
    return prep_list[prep];
}

/* Verb lookup */

#define M_OPT(tplate, arg, this) ( \
		   (tplate) == (int) OPT_ANY \
		|| ((tplate) == (int) OPT_NONE && (arg) == NOTHING) \
		|| ((tplate) == (int) OPT_THIS && (arg) == (this)) \
		)

static Program *
check_verbdefs(Objid start, const char *verbname, Parse_Info *pi,
	       int match_options)
{
    Verbdef    *v;

    pi->vloc = start;
    for (v = retrieve(start)->verbdefs; v; v = v->next) {
	if (verbcasecmp(v->name, verbname)) {
	    if (match_options
		   ? (M_OPT((v->perms >> VP_DOBJSHIFT) & VP_OBJMASK,
			    pi->dobj, pi->this)
		      && M_OPT((v->perms >> VP_IOBJSHIFT) & VP_OBJMASK,
			       pi->iobj, pi->this)
		      && (v->prep == AMBIGUOUS || v->prep == pi->prep))
		   : v->perms & (1 << VP_EXEC)) {
		pi->debug = v->perms & (1 << VP_DEBUG);
		pi->progr = v->owner;
		pi->verbname = strdup(v->name);
		return v->program;
	    }
	}
    }
    return 0;
}

static Verbdef *
check_verbdefs_simple(Objid start, const char *verbname)
{
    Verbdef    *v;

    for (v = retrieve(start)->verbdefs; v; v = v->next)
	if (verbcasecmp(v->name, verbname))
	    if (v->perms & (1 << VP_EXEC))
		return v;
    return 0;
}

Verbdef *
find_verb(Objid what, const char *verbname, Objid *vloc)
{
    Objid	oid;

    if (!retrieve(what))
	return 0;

    for (oid = what; oid != NOTHING; oid = retrieve(oid)->parent) {
	Verbdef *verbdef = check_verbdefs_simple(oid, verbname);

	if (verbdef) {
	    *vloc = oid;
	    return verbdef;
	}
    }
    return 0;
}

Program *
find_verb_on(Objid what, Objid this, const char *verbname,
	     Parse_Info *pi, int match_options)
{
    Objid	oid;

    if (!retrieve(what) || !retrieve(this))
	return 0;

    pi->this = this;
    for (oid = what; oid != NOTHING; oid = retrieve(oid)->parent) {
	Program *program = check_verbdefs(oid, verbname, pi, match_options);

	if (program)
	    return program;
    }
    return 0;
}

/* Command parsing */

Var
parse_into_wordlist(char *command)
{
    int		argc, i;
    char       *argv[MAXARGS];
    Var		args;

    parse_into_words(command, &argc, argv);
    args = new_list(argc);
    for (i = 1; i <= argc; i++) {
	args.v.list[i].type = TYPE_STR;
	args.v.list[i].v.str = strdup(argv[i - 1]);
    }
    return args;
}

int
parse_command(char *command, Parse_Info *pi)
{
    char       *argstr, *buf;
    const char *verb;
    int		argc;
    char       *argv[MAXARGS];
    int		pstart, pend, dlen;
    int		i;

    while (*command == ' ')
	command++;
    switch (*command) {
      case '"':
	verb = "say";
	goto finish_specials;
      case ':':
	verb = "emote";
	goto finish_specials;
      case ';':
	verb = "eval";
	goto finish_specials;
	
      finish_specials:
	argstr = command + 1;
	buf = (char *) mymalloc(strlen(argstr) + strlen(verb) + 2,
				M_STRING);
	strcpy(buf, verb);
	strcat(buf, " ");
	strcat(buf, argstr);
	break;
	
      default:
	buf = strdup(command);
	argstr = command;
	while (*argstr && *argstr != ' ')
	    argstr++;
	while (*argstr == ' ')
	    argstr++;
    }
    parse_into_words(buf, &argc, argv);

    if (argc == 0) {
	myfree(buf, M_STRING);
	return 0;
    }

    pi->argstr = strdup(argstr);
    pi->verb = strdup(argv[0]);

    pi->args = new_list(argc - 1);
    for (i = 1; i < argc; i++) {
	pi->args.v.list[i].type = TYPE_STR;
	pi->args.v.list[i].v.str = strdup(argv[i]);
    }

    /*
     * look for a preposition
     */
    if (argc > 1) {
	pi->prep = prep_find(argc - 1, argv + 1, &pstart, &pend);
	if (pi->prep == FAILED_MATCH) {
	    pi->prep = NOTHING;
	    pstart = argc;
	    pend = argc;
	} else {
	    pstart++;
	    pend++;
	}
    } else {
	pi->prep = NOTHING;
	pstart = argc;
	pend = argc;
    }

    /*
     * if there's a preposition,
     * find the iobj & dobj around it, if any
     */
    if (pi->prep != NOTHING) {
	pi->prepstr = build_string(pend - pstart + 1, argv + pstart);
	pi->iobjstr = build_string(argc - (pend + 1), argv + (pend + 1));
	pi->iobj = match_object(pi->player, pi->iobjstr);
    } else {
	pi->prepstr = strdup("");
	pi->iobjstr = strdup("");
	pi->iobj = NOTHING;
    }

    dlen = pstart - 1;
    if (dlen == 0) {
	pi->dobjstr = strdup("");
	pi->dobj = NOTHING;
    } else {
	pi->dobjstr = build_string(dlen, argv + 1);
	pi->dobj = match_object(pi->player, pi->dobjstr);
    }

    myfree(buf, M_STRING);

    return 1;
}

/* $Log: parse_command.c,v $
 * Revision 1.3  1992/08/10  17:35:34  pjames
 * Added find_verb() which doesn't fill in a Parse_Info, just returns
 * information about the verb.
 *
 * Revision 1.2  1992/07/21  00:04:53  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 * */
