char rcsid_property[] = "$Id: property.c,v 1.3 1992/08/10 17:41:48 pjames Exp $";

#include "execute.h"
#include "functions.h"
#include "list.h"
#include "log.h"
#include "match.h"
#include "property.h"
#include "storage.h"
#include "structures.h"
#include "unparse.h"
#include "utils.h"

int
count_properties(Objid oid)
{
    Propdef	*p;
    Object	*c;
    int		nprop = 0;

    for (c = retrieve(oid); c; c = retrieve(c->parent))
	for (p = c->propdefs; p; p = p->next)
	    nprop++;

    return nprop;
}

static Propdef *
new_propdef(const char *name)
{
    Propdef	*newprop;

    newprop = (Propdef *) mymalloc(sizeof(Propdef), M_PROPDEF);
    newprop->name = strdup(name);
    newprop->next = 0;
    return newprop;
}

static void
insert_prop(Objid what, int pos, Pval pv)
{
    Pval	*npval;
    Object	*o;
    int		i, nprops;

    o = retrieve(what);

    /*
     * allocate new property area
     */
    nprops = count_properties(what);
    npval = (Pval *) mymalloc(nprops * sizeof(Pval), M_PVAL);
    for (i = 0; i < pos; i++)
	npval[i] = o->propval[i];
    for (i = pos; i < nprops - 1; i++)
	npval[i+1] = o->propval[i];
    npval[pos] = pv;	/* assign new property */
    npval[pos].var = vardup(pv.var);
    if (npval[pos].perms & (1<<PP_CHOWN))
	npval[pos].owner = o->owner;

    if (o->propval)
	myfree((void *) o->propval, M_PVAL);
    o->propval = npval;
}

static void
remove_prop(Objid what, int pos)
{
    Pval	*npval;
    Object	*o;
    int		i, nprops;

    o = retrieve(what);
    free_var((o->propval)[pos].var);	/* free deleted property */

    nprops = count_properties(what);
    if (nprops) {
	npval = (Pval *) mymalloc(nprops * sizeof(Pval), M_PVAL);
	for (i = 0; i < pos; i++)
	    npval[i] = o->propval[i];
	for (i = pos; i < nprops; i++)
	    npval[i] = o->propval[i+1];
    } else
	npval = 0;

    if (o->propval)
	myfree((void *) o->propval, M_PVAL);
    o->propval = npval;
}

static void
insert_prop_recursively(Objid root, int root_pos, Pval pv)
{
    Objid	c;

    insert_prop(root, root_pos, pv);
    for (c = retrieve(root)->child; c != NOTHING; c = retrieve(c)->sibling) {
	int	new_prop_count = 0;
	Propdef *p;

	for (p = retrieve(c)->propdefs; p; p = p->next)
	    new_prop_count++;
	insert_prop_recursively(c, new_prop_count + root_pos, pv);
    }
}

/*
 * Add a new property to the given object (and all of its descendants).
 * Assume that permissions have already been checked, that the property name
 * is legal and that no property with this name already exists on the object.
 * Return the index of the new property.
 */
int
add_property(Objid owner, Objid oid, const char *propname, Var init, int perms)
{
    Object     *o = retrieve(oid);
    Propdef    *p = new_propdef(propname);
    Pval	pv;

    p->next = o->propdefs;
    o->propdefs = p;

    pv.var = init;
    pv.owner = owner;
    pv.perms = perms;

    insert_prop_recursively(oid, 0, pv);
    return 0;
}

static void
remove_prop_recursively(Objid root, int root_pos)
{
    Objid	c;

    remove_prop(root, root_pos);
    for (c = retrieve(root)->child; c != NOTHING; c = retrieve(c)->sibling) {
	int	new_prop_count = 0;
	Propdef *p;

	for (p = retrieve(c)->propdefs; p; p = p->next)
	    new_prop_count++;
	remove_prop_recursively(c, new_prop_count + root_pos);
    }
}

int
delete_property(Objid oid, const char *prop_name)
{
    int		pos = 0;
    Propdef    *p, *prev = 0;

    for (p = retrieve(oid)->propdefs; p; prev = p, p = p->next) {
	if (!mystrcasecmp(p->name, prop_name)) {
	    if (prev)
		prev->next = p->next;
	    else
		retrieve(oid)->propdefs = p->next;
	    if (p->name)
		myfree((void *) p->name, M_STRING);
	    myfree((void *) p, M_PROPDEF);
	    remove_prop_recursively(oid, pos);
	    return 1;
	}
	pos++;
    }
    return 0;
}

void
get_bi_prop(Objid obj, int pid, Var *ret)
{
    switch (pid) {
      case LOCATION:
	ret->type = TYPE_OBJ;
	ret->v.obj = retrieve(obj)->location;
	break;
      case OWNER:
	ret->type = TYPE_OBJ;
	ret->v.obj = retrieve(obj)->owner;
	break;
      case NAME:
	ret->type = TYPE_STR;
	ret->v.str = strdup(retrieve(obj)->name);
	break;
      case CONTENTS:
	{
	    Var	item, r;
	    Objid	oid;
	    int	i, count = 0;
	    
	    for (oid = retrieve(obj)->contents;
		 oid != NOTHING;
		 oid = retrieve(oid)->next)
		count++;
	    r = new_list(count);
	    item.type = TYPE_OBJ;
	    for (oid = retrieve(obj)->contents, i = 1;
		 oid != NOTHING;
		 oid = retrieve(oid)->next, i++) {
		item.v.obj = oid;
		r.v.list[i] = item;
	    }
	    *ret = r;
	}
	break;
      case FLAG - FLAG_PLAYER:
	ret->type = TYPE_NUM;
	ret->v.num = is_player(obj);
	break;
      case FLAG - FLAG_PROGRAMMER:
	ret->type = TYPE_NUM;
	ret->v.num = is_programmer(obj);
	break;
      case FLAG - FLAG_WIZARD:
	ret->type = TYPE_NUM;
	ret->v.num = is_wizard(obj);
	break;
      case FLAG - FLAG_READ:
	ret->type = TYPE_NUM;
	ret->v.num = retrieve(obj)->flags & (1 << FLAG_READ);
	break;
      case FLAG - FLAG_WRITE:
	ret->type = TYPE_NUM;
	ret->v.num = retrieve(obj)->flags & (1 << FLAG_WRITE);
	break;
      default:
	errlog("GET_BI_PROP: Unknown PID: %d\n", pid);
	ret->type = TYPE_NUM;
	ret->v.num = 0;
	break;
    }
}

enum error
get_prop(Objid obj, const char *name, Var *value, Objid progr)
{
    int		pid;
    
    if (!match_property(name, obj, &pid))
	return E_PROPNF;
    
    if (pid < 0) {
	get_bi_prop(obj, pid, value);
	return E_NONE;
    } else {
	Pval	*prop = retrieve(obj)->propval + pid;
	
	if (Prop_perm(progr, prop, PP_READ)) {
	    *value = vardup(prop->var);
	    return E_NONE;
	} else {
	    return E_PERM;
	}
    }
}

enum error
set_bi_prop(Objid obj, int pid, Var value, Objid progr)
{
    if (progr != retrieve(obj)->owner && !is_wizard(progr))
	return E_PERM;
    
    switch (pid) {
      case NAME:
	if (value.type != TYPE_STR)
	    return E_TYPE;
	if (is_player(obj) && !is_wizard(progr))
	    return E_PERM;
	if (retrieve(obj)->name)
	    myfree((void *) retrieve(obj)->name, M_STRING);
	retrieve(obj)->name = value.v.str;
	break;
      case OWNER:
	if (value.type != TYPE_OBJ)
	    return E_TYPE;
	if (!is_wizard(progr))
	    return E_PERM;
	if (!is_player(value.v.obj))
	    return E_INVARG;
	retrieve(obj)->owner = value.v.obj;
	break;
      case FLAG - FLAG_PROGRAMMER:
      case FLAG - FLAG_WIZARD:
	if (!is_wizard(progr))
	    return E_PERM;    
	if (!is_player(obj))
	    return E_INVARG;
	if (pid == FLAG - FLAG_WIZARD  &&  is_true(value)) {
	    log("WIZARDED: #%d by programmer #%d\n", obj, progr);
	    print_error_backtrace("Wizard bit set.", 1);
	}
	/* fall through to ... */
      case FLAG - FLAG_READ:
      case FLAG - FLAG_WRITE:
	{
	    int	flag = FLAG - pid;
	    
	    retrieve(obj)->flags &= ~(1 << flag);
	    if (is_true(value))
		retrieve(obj)->flags |= (1 << flag);
	}
	break;
      default:
	return E_PERM;
    }
    
    return E_NONE;
}

enum error
put_prop(Objid obj, const char *name, Var value, Objid progr)
{
    int		pid;
    
    if (!match_property(name, obj, &pid))
	return E_PROPNF;
    
    if (pid < 0) {
	if (progr != retrieve(obj)->owner && !is_wizard(progr)) {
	    return E_PERM;
	} else {
	    return set_bi_prop(obj, pid, vardup(value), progr);
	}
    } else {
	Pval   *prop = retrieve(obj)->propval + pid;
	
	if (Prop_perm(progr, prop, PP_WRITE)) {
	    free_var(prop->var);
	    prop->var = vardup(value);
	} else {
	    return E_PERM;
	}
    }
    return E_NONE;
}

/**** built in function support ****/

static int
is_prop_info(Var *list)
{
    return (list[0].v.num == 2
	    && list[1].type == TYPE_OBJ
	    && list[2].type == TYPE_STR);
}

static int
parse_prop_perms(const char *perm_str)
{
    int	new_perms = 0;
    
    while (*perm_str) {
	switch (*(perm_str++)) {
	  case 'r':
	  case 'R':
	    new_perms |= (1 << PP_READ);
	    break;
	  case 'w':
	  case 'W':
	    new_perms |= (1 << PP_WRITE);
	    break;
	  case 'c':
	  case 'C':
	    new_perms |= (1 << PP_CHOWN);
	    break;
	  default:
	    return -1;
	}
    }
    
    return new_perms;
}

/**** built in functions ****/

static package
bf_properties(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object) */
    Var r;
    Objid oid = arglist.v.list[1].v.obj;
    Object *o = retrieve(oid);
    
    free_var(arglist);
    
    if (!o)
	return make_error_pack(E_INVARG);
    else if (!Obj_perm(progr, oid, FLAG_READ)) 
	return make_error_pack(E_PERM);
    else {
	Propdef    *p;
	Var		prop;
	int		i, count = 0;
	
	for (p = o->propdefs; p; p = p->next)
	    count++;
	r = new_list(count);
	prop.type = TYPE_STR;
	for (i = 1, p = o->propdefs; p; p = p->next, i++) {
	    prop.v.str = strdup(p->name);
	    r.v.list[i] = prop;
	}
    }
    return make_var_pack(r);
}


static package
bf_prop_info(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, prop-name) */
    Var r;
    Objid oid = arglist.v.list[1].v.obj;
    Object *o;
    int pid;
    Propdef *match_result = match_property(arglist.v.list[2].v.str, oid, &pid);
    
    free_var(arglist);
    
    if (!(o = retrieve(oid)))
	return make_error_pack(E_INVARG);
    else 
	if (!match_result || pid < 0) 
	    return make_error_pack(E_PROPNF);
	else {
	    Pval   *pval = o->propval + pid;
	    int	i = 0;
	    char   *s;
	    
	    if (!Prop_perm(progr, pval, PP_READ)) 
		return make_error_pack(E_PERM);
	    else {
		r = new_list(2);
		r.v.list[1].type = TYPE_OBJ;
		r.v.list[1].v.obj = pval->owner;
		r.v.list[2].type = TYPE_STR;
		r.v.list[2].v.str = s = strdup("xxx");
		if (pval->perms & (1 << PP_READ)) s[i++] = 'r';
		if (pval->perms & (1 << PP_WRITE)) s[i++] = 'w';
		if (pval->perms & (1 << PP_CHOWN)) s[i++] = 'c';
		s[i] = '\0';
	    }
	}
    return make_var_pack(r);
}

static package
bf_set_prop_info(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, prop-name, {owner, perms}) */
    if (!is_prop_info(arglist.v.list[3].v.list)) {
	free_var(arglist);
	return make_error_pack(E_TYPE);
    } else if (!retrieve(arglist.v.list[1].v.obj)) {
	free_var(arglist);
	return make_error_pack(E_INVARG);
    } else {
	int		pid;
	
	if (!match_property(arglist.v.list[2].v.str, 
			    arglist.v.list[1].v.obj, &pid) || pid < 0) {
	    free_var(arglist);
	    return make_error_pack(E_PROPNF);
	} else {
	    Pval   *pval = retrieve(arglist.v.list[1].v.obj)->propval + pid;
	    Objid	new_owner = arglist.v.list[3].v.list[1].v.obj;
	    char   *perm_str = arglist.v.list[3].v.list[2].v.str;
	    int	new_perms = parse_prop_perms(perm_str);
	    
	    free_var(arglist);
	    
	    if ((!is_wizard(progr) && new_owner != pval->owner)
		|| !Prop_perm(progr, pval, PP_WRITE)) 
		return make_error_pack(E_PERM);
	    else if (new_perms < 0
		     || !retrieve(new_owner)
		     || !is_player(new_owner)) 
		return make_error_pack(E_INVARG);
	    else {
		pval->owner = new_owner;
		pval->perms = new_perms;
	    }
	}
    }
    return no_var_pack();
}

static int
property_defined_at_or_below(const char *pname, Objid oid)
{
    /* Return true iff some descendant of OID defines a property named PNAME */
    
    Objid	c;
    Propdef    *p;
    
    for (p = retrieve(oid)->propdefs; p; p = p->next)
	if (!mystrcasecmp(p->name, pname))
	    return 1;
    
    for (c = retrieve(oid)->child; c != NOTHING; c = retrieve(c)->sibling)
	if (property_defined_at_or_below(pname, c))
	    return 1;
    
    return 0;
}

static package
bf_add_prop(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, prop-name, initial-value, initial-info) */
    
    Objid	oid = arglist.v.list[1].v.obj;
    char       *pname = arglist.v.list[2].v.str;
    Var		value;
    Var	       *info = arglist.v.list[4].v.list;
    
    value = arglist.v.list[3];
    if (!is_prop_info(info)) {
	free_var(arglist);
	return make_error_pack(E_TYPE);
    } else {
	int	pid;
	Objid	new_owner = info[1].v.obj;
	int	perms = parse_prop_perms(info[2].v.str);
	
	if (!retrieve(oid)
	    || match_property(pname, oid, &pid)
	    || property_defined_at_or_below(pname, oid)
	    || perms < 0
	    || !retrieve(new_owner)
	    || !is_player(new_owner)) {
	    free_var(arglist);
	    return make_error_pack(E_INVARG); 
	} else if (!Obj_perm(progr, oid, FLAG_WRITE)
		   || (!is_wizard(progr) && (progr != new_owner))) {
	    free_var(arglist);
	    return make_error_pack(E_PERM);
	} else
	    (void) add_property(new_owner, oid, pname, value, perms);
    }
    free_var(arglist);
    return no_var_pack();
}

static package
bf_delete_prop(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, prop-name) */
    
    if (!retrieve(arglist.v.list[1].v.obj)) {
	free_var(arglist);
	return make_error_pack(E_INVARG);
    } else if (!Obj_perm(progr, arglist.v.list[1].v.obj, FLAG_WRITE)) {
	free_var(arglist);
	return make_error_pack(E_PERM);
    } else if (!delete_property(arglist.v.list[1].v.obj, 
				arglist.v.list[2].v.str)) {
	free_var(arglist);
	return make_error_pack(E_PROPNF);
    } 
    
    free_var(arglist);
    return no_var_pack();
}

void
register_bf_prop(void)
{ 
    (void) register_function("properties", 1, 1, bf_properties, TYPE_OBJ);
    (void) register_function("property_info", 2, 2, bf_prop_info, TYPE_OBJ, TYPE_STR);
    (void) register_function("set_property_info", 3, 3, bf_set_prop_info,
			     TYPE_OBJ, TYPE_STR, TYPE_LIST);
    (void) register_function("add_property", 4, 4, bf_add_prop, TYPE_OBJ, TYPE_STR,
			     TYPE_ANY, TYPE_LIST);
    (void) register_function("delete_property", 2, 2, bf_delete_prop,
			     TYPE_OBJ, TYPE_STR);
}

/* $Log: property.c,v $
 * Revision 1.3  1992/08/10  17:41:48  pjames
 * Moved get_bi_prop, get_prop, set_bi_prop, and put_prop from execute.c.
 *
 * Revision 1.2  1992/07/21  00:06:01  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 * */
