char rcsid_timers[] = "$Id: timers.c,v 1.3 1992/07/30 00:42:45 pavel Exp $";

#include "my-signal.h"
#include "my-stdlib.h"
#include "my-sys-time.h"
#include "my-time.h"
#include "my-unistd.h"

#include "timers.h"

typedef struct Timer_Entry	Timer_Entry;
struct Timer_Entry {
    Timer_Entry *next;
    time_t	when;
    Timer_Proc	proc;
    Timer_Data	data;
    Timer_ID	id;
};

static Timer_Entry     *active_timers = 0;
static Timer_Entry     *free_timers = 0;
static Timer_Entry     *virtual_timer = 0;
static Timer_ID		next_id = 0;

static Timer_Entry *
allocate_timer(void)
{
    if (free_timers) {
	Timer_Entry    *this = free_timers;

	free_timers = this->next;
	return this;
    } else
	return (Timer_Entry *) malloc(sizeof(Timer_Entry));
}

static void
free_timer(Timer_Entry *this)
{
    this->next = free_timers;
    free_timers = this;
}

static void		restart_timers(void);

static void
wakeup_call(int signo)
{
    Timer_Entry	*this = active_timers;
    Timer_Proc	proc = this->proc;
    Timer_ID	id = this->id;
    Timer_Data	data = this->data;

    active_timers = active_timers->next;
    free_timer(this);
    restart_timers();
    if (proc)
	(*proc)(id, data);
}

static void
virtual_wakeup_call(int signo)
{
    Timer_Entry	*this = virtual_timer;
    Timer_Proc	proc = this->proc;
    Timer_ID	id = this->id;
    Timer_Data	data = this->data;

    virtual_timer = 0;
    free_timer(this);
    if (proc)
	(*proc)(id, data);
}    

static void
stop_timers()
{
    alarm(0);
    signal(SIGALRM, SIG_IGN);
    signal(SIGALRM, wakeup_call);

#ifdef ITIMER_VIRTUAL
    {
	struct itimerval	itimer, oitimer;

	itimer.it_value.tv_sec = 0;
	itimer.it_value.tv_usec = 0;
	itimer.it_interval.tv_sec = 0;
	itimer.it_interval.tv_usec = 0;

	setitimer(ITIMER_VIRTUAL, &itimer, &oitimer);
	signal(SIGVTALRM, SIG_IGN);
	signal(SIGVTALRM, virtual_wakeup_call);
	if (virtual_timer)
	    virtual_timer->when = oitimer.it_value.tv_sec;
    }
#endif
}

static void
restart_timers()
{
    if (active_timers) {
	time_t	now = time(0);

	signal(SIGALRM, wakeup_call);

	if (now < active_timers->when)	/* first timer is in the future */
	    alarm(active_timers->when - now);
	else
	    kill(getpid(), SIGALRM);	       	/* we're already late... */
    }

#ifdef ITIMER_VIRTUAL

    if (virtual_timer) {
	signal(SIGVTALRM, virtual_wakeup_call);

	if (virtual_timer->when > 0) {
	    struct itimerval	itimer;

	    itimer.it_value.tv_sec = virtual_timer->when;
	    itimer.it_value.tv_usec = 0;
	    itimer.it_interval.tv_sec = 0;
	    itimer.it_interval.tv_usec = 0;

	    setitimer(ITIMER_VIRTUAL, &itimer, 0);
	} else
	    kill(getpid(), SIGVTALRM);
    }

#endif
}

Timer_ID
set_timer(unsigned seconds, Timer_Proc proc, Timer_Data data)
{
    Timer_Entry	*this = allocate_timer();
    Timer_Entry **t;

    this->id = next_id++;
    this->when = time(0) + seconds;
    this->proc = proc;
    this->data = data;
    
    stop_timers();

    t = &active_timers;
    while (*t  &&  this->when >= (*t)->when)
	t = &((*t)->next);
    this->next = *t;
    *t = this;

    restart_timers();

    return this->id;
}

Timer_ID
set_virtual_timer(unsigned seconds, Timer_Proc proc, Timer_Data data)
{
#ifdef ITIMER_VIRTUAL

    if (virtual_timer)
	return -1;

    stop_timers();

    virtual_timer = allocate_timer();
    virtual_timer->id = next_id++;
    virtual_timer->when = seconds;
    virtual_timer->proc = proc;
    virtual_timer->data = data;

    restart_timers();

    return virtual_timer->id;

#else /* !ITIMER_VIRTUAL */

    return set_timer(seconds, proc, data);

#endif
}

int
virtual_timer_available()
{
#ifdef ITIMER_VIRTUAL
    return 1;
#else
    return 0;
#endif
}

unsigned
timer_wakeup_interval(Timer_ID id)
{
    Timer_Entry	*t;

#ifdef ITIMER_VIRTUAL

    if (virtual_timer  &&  virtual_timer->id == id) {
	struct itimerval	itimer;

	getitimer(ITIMER_VIRTUAL, &itimer);
	return itimer.it_value.tv_sec;
    }

#endif

    for (t = active_timers; t; t = t->next)
	if (t->id == id)
	    return t->when - time(0);;

    return 0;
}

void
timer_sleep(unsigned seconds)
{
    set_timer(seconds, 0, 0);
    pause();
}

int
cancel_timer(Timer_ID id)
{
    Timer_Entry	**t = &active_timers;
    int		found = 0;

    stop_timers();

    if (virtual_timer  &&  virtual_timer->id == id) {
	free(virtual_timer);
	virtual_timer = 0;
	found = 1;
    } else {
	while (*t) {
	    if ((*t)->id == id) {
		Timer_Entry	*tt = *t;
		
		*t = tt->next;
		found = 1;
		free(tt);
		break;
	    }
	    t = &((*t)->next);
	}
    }

    restart_timers();

    return found;
}

void
reenable_timers(void)
{
#ifdef USE_POSIX
    sigset_t	sigs;

    (void) sigemptyset(&sigs);
    (void) sigaddset(&sigs, SIGALRM);
    (void) sigprocmask(SIG_UNBLOCK, &sigs, NULL);
#else
#ifdef MIPS_SVR3
    sigrelse(SIGALRM);			/* restore previous signal action */
#else  /* BSD signals */
    int		old_mask = sigsetmask(-1); /* block everything, get old mask */

    old_mask &= ~sigmask(SIGALRM);	/* clear blocked bit for SIGALRM */
    sigsetmask(old_mask);		/* reset the signal mask */
#endif
#endif
}

/* $Log: timers.c,v $
 * Revision 1.3  1992/07/30  00:42:45  pavel
 * Add support for compiling on RISC/os 4.52 and NonStop-UX A22.
 *
 * Revision 1.2  1992/07/21  00:07:26  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 * */
